unit CDRIP;
{CD ripper component based on CDEX ripping engine

 Version 1.1}

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs, CDRip_dll,
  LameCoder;

type ECDRIPError = class(Exception);

type TCDROMName = array[0..255] of char;

type
    TTrack = class(TObject)
       TrackNum : Integer;
       BegTrack : LongInt;
       EndTrack : LongInt;
    end;
type
  TProgressEvent = procedure(Sender: TObject; PercentComplete : Word) of object;
  TProcessTrackEvent = procedure(Sender: TObject; TrackNum : Word) of object;

type
  TDriveTableClass = class(TPersistent)
  private
     FDTable : TDriveTable;
  published
     property DriveType  : TDriveType read FDTable.DriveType write FDTable.DriveType;
     property ReadMethod : TReadMethod read FDTable.ReadMethod write FDTable.ReadMethod;
     property SetSpeed   : TSetSpeed read FDTable.SetSpeed write FDTable.SetSpeed;
     property Endian     : TEndian read FDTable.Endian write FDTable.Endian;
     property EnableMode : TEnableMode read FDTable.EnableMode write FDTable.EnableMode;
     property Density   : LongInt read FDTable.nDensity write FDTable.nDensity;
     property Atapi     : BOOL read FDTable.bAtapi write FDTable.bAtapi;
  end;

type
  TCDRIP = class(TComponent)
  private
    { Private declarations }
    FCDRomNum     : LongInt;
    FDriveTable   : TDriveTableClass;
    FOutputFormat : TOutputFormat;
    FCDRomName    : TCDROMName;
    FLameCoder    : TLameCoder;
    FTocNum       : LongInt;
    FTracks       : TStrings;
    FJitters      : TStrings;
    FJittersNum   : LongInt;
    FJitter       : boolean;
    FEncode       : boolean;
    FWriteWAV     : boolean;
    FCancelRip    : boolean;
    FSpeed        : integer;
    FOnBeginTrack    : TProcessTrackEvent;
    FOnEndTrack      : TProcessTrackEvent;
    FOnProgress     : TProgressEvent;
    FVersion       : string;
    procedure SetActiveCD(Value : LongInt);
    procedure SetOutputFormat(Value : TOutputFormat);
    procedure SetLameCoder(Value : TLameCoder);
    procedure SetLameFormat;
    procedure SetJitterCorrection(Value : boolean);
  protected
    { Protected declarations }
  public
    { Public declarations }
     constructor Create(AOwner: TComponent); override;
     destructor Destroy; override;
     function IsCDROMReady : boolean;
     function SetNewCDROMParams : boolean;
     procedure RipTrack(TrackNum : LongInt; FileName : string);
     procedure CancelRip;
     property CDRomName : TCDRomName read FCDRomName write FCDRomName;
     property TrackNum    : LongInt read FTOCNum;
     property Tracks    : TStrings read FTracks;
     property Jitters   : TStrings read FJitters;
     property JittersNum : LongInt read FJittersNum;
  published
    { Published declarations }
     property Coder : TLameCoder read FLameCoder write SetLameCoder;
     property CDRomNum : LongInt read FCDRomNum write SetActiveCD;
     property DriveTable : TDriveTableClass read FDriveTable write FDriveTable;
     property OutputFormat : TOutputFormat read FOutputFormat write SetOutputFormat default STEREO44100;
     property Encode : boolean read FEncode write FEncode default false;
     property WriteWAV       : boolean read FWriteWAV write FWriteWAV default false;
     property JitterCorrect  : boolean read FJitter write SetJitterCorrection default true;
     property OnBeginTrack   : TProcessTrackEvent read FOnBeginTrack    write FOnBeginTrack;
     property OnEndTrack     : TProcessTrackEvent read FOnEndTrack      write FOnEndTrack;
     property OnProgress     : TProgressEvent     read FOnProgress     write FOnProgress;
     property Version        : string read FVersion write FVersion;
     property Speed          : integer read FSpeed write FSpeed;
  end;

procedure Register;
{$R CDRip.res}
implementation

var
   FParams : TCDROMParams;

function LeadingZero2(w: Word): String;
var s: String;
begin
  Str(w:0, s);
  LeadingZero2 := Copy('00', 1, 2 - Length(s)) + s;
end;

constructor TCDRIP.Create(AOwner: TComponent);
var i : LongInt;
    S : string;
    L1, L2 : LongInt;
begin
     inherited Create(AOwner);
     try
     FTracks:=TStringList.Create;
     FJitters:=TStringList.Create;
     L1:=CR_GetCDRipVersion;
     FVersion:=IntToStr(L1);
     if CR_Init=CDEX_OK then begin
       if CR_GetCDROMParameters(FParams)=CDEX_OK then begin
         FCDRomName:=TCDROMName(FParams.lpszCDROMID);
         FCDRomNum:=CR_GetActiveCDROM;
         FDriveTable:=TDriveTableClass.Create;
         FDriveTable.FDTable:=FParams.DriveTable;
         FOutputFormat:=FParams.nOutputFormat;
         FSpeed:=FParams.nSpeed;
         FJitter:=true;
         FParams.bJitterCorrection:=LongBool(FJitter);
        end;
        if CR_ReadTOC = CDEX_OK then FTOCNum:=CR_GetNumTocEntries;
        for i:=0 to FTOCNum-1 do begin
            S:='Track'+LeadingZero2(i+1);
            L1:=CR_GetTOCEntry(i);
            L2:=CR_GetTOCEntry(i+1);
            FTracks.AddObject(S,TTrack.Create);
            TTrack(FTracks.Objects[i]).TrackNum:=i+1;
            TTrack(FTracks.Objects[i]).BegTrack:=L1;
            TTrack(FTracks.Objects[i]).EndTrack:=L2-1;
        end;
       end;
       FJittersNum:=0;
     except
        raise ECDRIPError.Create('Initialization failure');
     end;
end;


destructor TCDRIP.Destroy;
begin
     FTracks.Free;
     FJitters.Free;
     FDriveTable.Free;
     FLameCoder:=nil;
     inherited Destroy;
end;

procedure TCDRIP.SetActiveCD(Value : LongInt);
var i: Integer;
    S : string;
    L1, L2 : LongInt;
begin
     if FCDRomNum<>Value then
     try
        CR_SetActiveCDROM(Value);
        if CR_GetCDROMParameters(FParams)=CDEX_OK then begin
           FCDRomName:=TCDROMName(FParams.lpszCDROMID);
           FDriveTable.FDTable:=FParams.DriveTable;
           FOutputFormat:=FParams.nOutputFormat;
           FSpeed:=FParams.nSpeed;
           FCDRomNum:=Value;
        end;
        if CR_ReadTOC = CDEX_OK then FTOCNum:=CR_GetNumTocEntries;
        for i:=0 to FTOCNum-1 do begin
            S:='Track'+LeadingZero2(i+1);
            L1:=CR_GetTOCEntry(i);
            L2:=CR_GetTOCEntry(i+1);
            FTracks.AddObject(S,TTrack.Create);
            TTrack(FTracks.Objects[i]).TrackNum:=i+1;
            TTrack(FTracks.Objects[i]).BegTrack:=L1;
            TTrack(FTracks.Objects[i]).EndTrack:=L2-1;
        end;
     except
        raise ECDRIPError.Create('Initialization failure');
     end;
end;


procedure TCDRIP.SetJitterCorrection(Value : boolean);
begin
     if FJitter<>Value then begin
        FJitter:=Value;
        FParams.bJitterCorrection:=LongBool(Value);
     end;
end;

procedure TCDRIP.SetLameFormat;
begin
     if FLameCoder<>nil then
     case FOutputFormat of
                STEREO44100 : begin
                                   FLameCoder.Mode       :=STEREO;
                                   FLameCoder.SampleRate :=44100;
                              end;
                MONO44100 :   begin
                                   FLameCoder.Mode       :=MONO;
                                   FLameCoder.SampleRate :=44100;
                              end;
                STEREO22050 : begin
                                   FLameCoder.Mode       :=STEREO;
                                   FLameCoder.SampleRate :=22050;
                              end;
                MONO22050 :   begin
                                   FLameCoder.Mode       :=MONO;
                                   FLameCoder.SampleRate :=22050;
                              end;
                STEREO11025 : begin
                                   FLameCoder.Mode       :=STEREO;
                                   FLameCoder.SampleRate :=11025;
                              end;
                MONO11025 :   begin
                                   FLameCoder.Mode       :=MONO;
                                   FLameCoder.SampleRate :=11025;
                              end;
                end;
end;
procedure TCDRIP.SetOutputFormat(Value : TOutputFormat);
begin
     if FOutputFormat<>Value then
     FParams.nOutputFormat:=Value;
     FOutputFormat:=Value;
     SetLameFormat;
end;

function TCDRIP.IsCDROMReady : boolean;
begin
     Result:=CR_IsUnitReady;
end;

function TCDRIP.SetNewCDROMParams : boolean;
begin
     try
        if CR_SetCDROMParameters(@FParams)=CDEX_OK
        then Result:=true else Result:=false;
     except
        raise ECDRIPError.Create('Error setting parameters');
        Result:=false;
     end;
end;

procedure TCDRIP.SetLameCoder(Value : TLameCoder);
begin
     FLameCoder:=Value;
     SetLameFormat;
end;

procedure TCDRIP.RipTrack(TrackNum : LongInt; FileName : String);
var Track : TTrack;
    Buffer,CBuffer,LameInBuffer,LameOutBuffer : pointer;
    BufSize,LameInBufSize : LongInt;
    CDStream : TFileStream;
    WaveHeader : TWaveHeader;
    Duration : Real;
    Ch, SR : integer;
    Data : LongInt;
    ERR : CDEX_ERR;
    BytesRipped,BytesEncoded : LongWord;
    StartJitter,EndJitter : LongWord;
    PrComplete : Integer;
    i : integer;
    NumBufs : integer;
    BytesToWrite,BytesWritten : LongWord;

begin
     if not SetNewCDROMParams then begin
        raise ECDRIPError.Create('Invalid CD-ROM parameters');
        Exit;
     end;
     if Assigned(FOnBeginTrack) then FOnBeginTrack(Self,TrackNum);
     FCancelRip:=false;

     Track:=TTrack(FTracks.Objects[TrackNum-1]);
     if CR_OpenRipper(@BufSize,Track.BegTrack,Track.EndTrack) = CDEX_OK then begin
        if FEncode then begin
           FileName:=ChangeFileExt(FileName,'.mp3');
           GetMem(Buffer,BufSize+FLameCoder.NumSamples*2);
        end else
           GetMem(Buffer,BufSize);

        CDStream:=TFileStream.Create(FileName,fmCreate);

        case FOutputFormat of
                STEREO44100 : begin
                                   Ch:=2;
                                   SR:=44100;
                              end;
                MONO44100 : begin
                                   Ch:=1;
                                   SR:=44100;
                              end;
                STEREO22050 : begin
                                   Ch:=2;
                                   SR:=22050;
                              end;
                MONO22050 : begin
                                   Ch:=1;
                                   SR:=22050;
                              end;
                STEREO11025 : begin
                                   Ch:=2;
                                   SR:=11025;
                              end;
                MONO11025 : begin
                                   Ch:=1;
                                   SR:=11025;
                              end;
        end;

        if FWriteWAV and (not FEncode) then begin
           Duration:=(Track.EndTrack-Track.BegTrack) / 75;

           Data:=Round((Ch*SR*2* Duration));

           with WaveHeader do begin
                Marker1:='RIFF';
                BytesFollowing:=Data+36;
                Marker2:='WAVE';
                Marker3:='fmt ';
                Fixed1:=16;
                FormatTag:=1;
                Channels:=Ch;
                SampleRate:=SR;
                BytesPerSecond:=Ch*SR*2;
                BytesPerSample:=Ch*2;
                BitsPerSample:=16;
                Marker4:='data';
                DataBytes:=Data;
           end;
           CDStream.Write(WaveHeader,SizeOf(WaveHeader));
        end;

        if FEncode and (FLameCoder<>nil) then begin
           FLameCoder.PrepareCoder;
           GetMem(LameOutBuffer,FLameCoder.BufSize);
           LameInBufSize:=FLameCoder.NumSamples*Ch;
           GetMem(LameInBuffer,LameInBufSize);
        end;

        BytesToWrite:=0;
        repeat

             if FCancelRip then
             if FEncode and (FLameCoder<>nil) then begin
                FLameCoder.UnPrepareCoder(LameOutBuffer,BytesEncoded);
                CDStream.Write(LameOutBuffer^,BytesEncoded);
                FLameCoder.CloseCoder;
                Exit;
             end else begin
                CR_CloseRipper;
                CDStream.Free;
                FreeMem(Buffer);
                Exit;
             end;

             ERR:=CR_RipChunk(PChar(Buffer)+BytesToWrite,BytesRipped);

             if ERR=CDEX_OK then begin
{Onfly encoding }
                if FEncode and (FLameCoder<>nil) then begin
                    BytesRipped:=BytesRipped+BytesToWrite;
                    NumBufs:=BytesRipped div LameInBufSize;
                    CBuffer:=Buffer;

                    for i:=1 to NumBufs do begin
                        BytesToWrite:=LameInBufSize;
                        Move(CBuffer^,LameInBuffer^,BytesToWrite);
                        PChar(CBuffer):=PChar(CBuffer)+BytesToWrite;
                        FLameCoder.EncodeBuffer(LameInBuffer,LameOutBuffer,BytesEncoded);
                        BytesWritten:=CDStream.Write(LameOutBuffer^,BytesEncoded);
                        if BytesWritten<>BytesEncoded then
                           raise ECDRIPError.Create('Write error');

                    end;

                        BytesToWrite:=BytesRipped-NumBufs*LameInBufSize;
                        if BytesToWrite>0 then

                        Move(CBuffer^,Buffer^,BytesToWrite);

                end;

{No encoding}
                if not FEncode then begin
                    BytesWritten:=CDStream.Write(Buffer^,BytesRipped);
                    if BytesWritten<>BytesRipped then
                       raise ECDRIPError.Create('Write error');
                end;

             end;

             if ERR=CDEX_JITTER_ERROR then begin
                CR_GetLastJitterErrorPosition(StartJitter,EndJitter);
                FJitters.Add(IntToStr(StartJitter)+' '+IntToStr(EndJitter));
             end;

             PrComplete:=CR_GetPercentCompleted;
        if Assigned(FOnProgress) then FOnProgress(Self,PrComplete);
             Application.ProcessMessages;

        until ERR=CDEX_RIPPING_DONE;

        if FEncode and (FLameCoder<>nil) then begin
           FLameCoder.UnPrepareCoder(LameOutBuffer,BytesEncoded);
           BytesWritten:=CDStream.Write(LameOutBuffer^,BytesEncoded);
           if BytesWritten<>BytesEncoded then
              raise ECDRIPError.Create('Write error');
           FLameCoder.CloseCoder;
        end;

        FJittersNum:=CR_GetNumberOfJitterErrors;
        CR_CloseRipper;
        FreeMem(Buffer);
        CDStream.Free;

        if FJitters.Count<>0 then FJitters.SaveToFile(ChangeFileExt(FileName,'jitt'));
        if Assigned(FOnEndTrack) then FOnEndTrack(Self,TrackNum);
   end;
end;

procedure TCDRIP.CancelRip;
begin
     FCancelRip:=true;
end;

procedure Register;
begin
  RegisterComponents('Lame', [TCDRIP]);
end;

end.
